package mpi.eudico.client.annotator.commands;

import java.util.ArrayList;
import java.util.List;

import mpi.eudico.client.annotator.imports.MergeUtil;
import mpi.eudico.client.annotator.util.ClientLogger;
import mpi.eudico.client.annotator.util.FileUtility;
import mpi.eudico.server.corpora.clomimpl.abstr.ParseException;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.abstr.TranscriptionImpl;
import mpi.eudico.server.corpora.clomimpl.dobes.EAFSkeletonParser;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;
import mpi.eudico.util.ControlledVocabulary;

/**
 * Import tiers (if necessary with Linguistic Types and ControlledVocabularies)
 * from an eaf or etf into a transcription. 
 */
public class ImportTiersCommand implements UndoableCommand, ClientLogger {
    private String commandName;
    
    // receiver
    private TranscriptionImpl transcription;
    
    private List<TierImpl> tiersAdded;
    private List<LinguisticType> typesAdded;
    private List<ControlledVocabulary> cvsAdded;
    
    public ImportTiersCommand(String name) {
        commandName = name;
    }
    
    /**
     * @see mpi.eudico.client.annotator.commands.UndoableCommand#undo()
     */
    @Override
	public void undo() {
        if(transcription == null) {
            LOG.warning("The transcription is null.");
            return;
        }
        if (tiersAdded == null) {
            LOG.warning("No tiers have been added.");
            return;
        }
        for (TierImpl t : tiersAdded) {
            transcription.removeTier(t);
        }
        for (LinguisticType lt :  typesAdded) {
            transcription.removeLinguisticType(lt);
        }
        for (ControlledVocabulary cv : cvsAdded) {
            transcription.removeControlledVocabulary(cv);
        }
    }

    /**
     * @see mpi.eudico.client.annotator.commands.UndoableCommand#redo()
     */
    @Override
	public void redo() {
        if(transcription == null) {
            LOG.warning("The transcription is null.");
            return;
        }
        if (tiersAdded == null) {
            LOG.warning("No tiers can be added.");
            return;
        }
        for (ControlledVocabulary cv : cvsAdded) {
            transcription.addControlledVocabulary(cv);
        }       
        for (LinguisticType lt : typesAdded) {
            transcription.addLinguisticType(lt);
        }
        for (TierImpl t : tiersAdded) {
            transcription.addTier(t);
        }
    }

    /**
     * <b>Note: </b>it is assumed the types and order of the arguments are
     * correct.
     *
     * @param receiver the transcription
     * @param arguments the arguments:  <ul><li>arg[0] = the fileName of an eaf
     *        or etf file (String)</li> </ul>
     */
    @Override
	public void execute(Object receiver, Object[] arguments) {
        transcription = (TranscriptionImpl) receiver;
        String fileName = (String) arguments[0];
        if (fileName == null) {
            LOG.warning("The filename is null");
            return; // report??
        }
        
        fileName = FileUtility.pathToURLString(fileName).substring(5);
        
        TranscriptionImpl srcTrans = null;
        try {
            EAFSkeletonParser parser = new EAFSkeletonParser(fileName);
            parser.parse();
            List<TierImpl> tiers = parser.getTiers();
            List<String> tierOrder = parser.getTierOrder();
            List<LinguisticType> types = parser.getLinguisticTypes();
            List<ControlledVocabulary> cvs = parser.getControlledVocabularies();
            
            srcTrans = new TranscriptionImpl();
            //srcTrans.setNotifying(false);
            srcTrans.setLinguisticTypes(new ArrayList<LinguisticType>(types));

            String tierName;
            TierImpl tier;
            for (int i = 0; i < tierOrder.size(); i++) {
                tierName = tierOrder.get(i);
                for (int j = 0; j < tiers.size(); j++) {
                	tier = tiers.get(j);
                	if (tierName.equals(tier.getName())) {
                		srcTrans.addTier(tier);	
                	}
                }
            }
            
            for (ControlledVocabulary cv : cvs) {
                srcTrans.addControlledVocabulary(cv);
            }
            
        } catch (ParseException pe) {
            LOG.warning(pe.getMessage());
            pe.printStackTrace();
            return;
        }
        // store current tiers types, cvs
        //transcription.setNotifying(false);
        
        List<TierImpl> currentTiers = new ArrayList<TierImpl>(transcription.getTiers());
        List<LinguisticType> currentTypes = new ArrayList<LinguisticType>(transcription.getLinguisticTypes());
        List<ControlledVocabulary> currentCvs = new ArrayList<ControlledVocabulary>(transcription.getControlledVocabularies());
        MergeUtil mergeUtil = new MergeUtil();
        List<TierImpl> tiersAddable = mergeUtil.getAddableTiers(srcTrans, transcription, null);
        if (tiersAddable == null || tiersAddable.size() == 0) {
            LOG.warning("There are no tiers that can be imported");
            transcription.setNotifying(true);
            return;
        }
        tiersAddable = mergeUtil.sortTiers(tiersAddable);
        mergeUtil.addTiersTypesAndCVs(srcTrans, transcription, tiersAddable);
        //store the tiers, types and cvs that have been added, for undo/redo
        //transcription.setNotifying(true);
        
        tiersAdded = new ArrayList<TierImpl>();
        typesAdded = new ArrayList<LinguisticType>();
        cvsAdded = new ArrayList<ControlledVocabulary>();
        
        for (TierImpl t : transcription.getTiers()) {
            if (!currentTiers.contains(t)) {
                tiersAdded.add(t);
            }
        }
        for (LinguisticType lt : transcription.getLinguisticTypes()) {
            if (!currentTypes.contains(lt)) {
                typesAdded.add(lt);
            }
        }
        for (ControlledVocabulary cv : transcription.getControlledVocabularies()) {
            if (!currentCvs.contains(cv)) {
                cvsAdded.add(cv);
            }
        }
    }

    /**
     * Returns the name of the command
     *
     * @return the name of the command
     */
    @Override
	public String getName() {
        return commandName;
    }


}
