package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.Component;
import java.awt.Dimension;
import java.awt.FlowLayout;
import java.awt.LayoutManager;

import javax.swing.JComponent;

import mpi.eudico.client.annotator.interlinear.edit.SuggestionComponent.MouseAndMotionListener;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionViewerModel;
import mpi.eudico.client.annotator.interlinear.edit.render.IGTViewerRenderInfo;
import mpi.eudico.client.annotator.interlinear.edit.render.SuggestionFlowLayout;

/**
 * A component that visualizes suggestion sets and allows to select one of the suggestions.
 * 
 * @author Han Sloetjes
 *
 */
@SuppressWarnings("serial")
public class SuggestionSetSelector extends JComponent {
	private IGTSuggestionViewerModel viewerModel;
	private IGTViewerRenderInfo viewerRenderInfo;
	
	private int selectorWidth;
	private int selectorHeight;
		
	/**
	 * Constructor accepting suggestions sets in a viewer model as parameter.
	 * 
	 * @param viewerModel the model containing the suggestion sets encapsulated in a igt model
	 * @param viewerRenderInfo contains information for the renderer
	 */
	public SuggestionSetSelector(IGTSuggestionViewerModel viewerModel, IGTViewerRenderInfo viewerRenderInfo) {
		super();
		this.viewerModel = viewerModel;
		this.viewerRenderInfo = viewerRenderInfo;
		
//		initComponents();
	}
	
	/**
	 * Call init component after the component is visible, otherwise somehow it
	 * doesn't know its contents properly... huh?
	 * <p>
	 * @see #initComponents()
	 */
	public void init() {
		initComponents();
	}

	/**
	 * Calculates the total height and width of all suggestion sets.
	 */
	private void initComponents() {
		if (viewerModel != null) {
			selectorWidth = viewerRenderInfo.headerWidth;
			selectorHeight = 0;
			
			final int margin = viewerModel.getRenderInfo().suggestionMargin;
			setLayout(new SuggestionFlowLayout(FlowLayout.LEFT, margin, margin));
			
			//setDebugGraphicsOptions(DebugGraphics.FLASH_OPTION);
			//setOpaque(false);	 // TESTING: where is all the flashing coming from??
			
			final int rowCount = viewerModel.getRowCount();

			for (int i = 0; i < rowCount; i++) {
				SuggestionComponent comp = new SuggestionComponent(i, viewerRenderInfo, viewerModel);
				add(comp);
			}
			validate();
						
		} else {
			selectorWidth = 150;
			selectorHeight = 100;
			setSize(selectorWidth, selectorHeight);
			setPreferredSize(new Dimension(selectorWidth, selectorHeight));
		}
	}

	/**
	 * Returns the suggestion viewer model.
	 * 
	 * @return the suggestion viewer model
	 */
	public IGTSuggestionViewerModel getModel() {
		return viewerModel;
	}

	/* (non-Javadoc)
	 * @see javax.swing.JComponent#getPreferredSize()
	 */
	@Override
	public Dimension getPreferredSize() {
		LayoutManager layout = getLayout();
		if (layout instanceof SuggestionFlowLayout) {
			return ((SuggestionFlowLayout)layout).getSize();
		}
		if (selectorHeight == 0 || selectorWidth == 0) {
			return new Dimension(260, 80);
		}
		
		return new Dimension(selectorWidth, selectorHeight);
	}
	
	/**
	 * Get the height of a single SuggestionComponent
	 * plus the vertical spacing.
	 * Hopefully they are all the same height.
	 */
	public int getUnitHeight() {
		if (getComponentCount() > 0) {
			return getComponent(0).getHeight() + viewerModel.getRenderInfo().suggestionMargin;
		}
		
		return 80;	// some random default
	}
	
	/**
	 * Returns the index of the model displayed at the given location.
	 * 
	 * @param x the x coordinate
	 * @param y the y coordinate
	 * @return the index of the model, or -1 if none
	 */
	public int getSuggestionIndexAtPoint(int x, int y) {
		Component comp = getComponentAt(x, y);
		// Does this handle the JScollPane correctly???
		
		if (comp instanceof SuggestionComponent) {
			SuggestionComponent sugg = (SuggestionComponent)comp;
			
			return sugg.getIndex();
		}
		return -1;
	}
	
	public void workAroundTooltipBug(MouseAndMotionListener mml) {
		Component[] children = getComponents();
		
		for (Component c : children) {
			if (c instanceof SuggestionComponent) {
				SuggestionComponent sugg = (SuggestionComponent)c;
				
				sugg.workAroundToolTipBug(mml);
			}
		}
	}
	
	public void removeWorkAroundToolTipBug(MouseAndMotionListener mml) {
		Component[] children = getComponents();
		
		for (Component c : children) {
			if (c instanceof SuggestionComponent) {
				SuggestionComponent sugg = (SuggestionComponent)c;
				
				sugg.removeWorkAroundToolTipBug(mml);
			}
		}
	}
	
	/** for setting a breakpoint */
//	@Override
//	public void update(Graphics g) {
//		super.update(g);
//	}
}
