package mpi.eudico.client.annotator.interlinear.edit;

import java.awt.Component;
import java.awt.Container;
import java.awt.Dimension;
import java.awt.Frame;
import java.awt.Insets;
import java.awt.Point;
import java.awt.Rectangle;
import java.awt.Window;
import java.awt.event.ComponentEvent;
import java.awt.event.ComponentListener;
import java.awt.event.KeyEvent;
import java.awt.event.KeyListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.awt.event.MouseMotionListener;
import java.util.logging.Level;

import javax.swing.DebugGraphics;
import javax.swing.JComponent;
import javax.swing.JFrame;
import javax.swing.JLabel;
import javax.swing.JScrollBar;
import javax.swing.JScrollPane;
import javax.swing.RepaintManager;
import javax.swing.UIManager;

import mpi.eudico.client.annotator.interlinear.edit.SuggestionComponent.MouseAndMotionListener;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionListener;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionViewerModel;
import mpi.eudico.client.annotator.util.ClientLogger;

/**
 * A window class that can show suggestions in a kind of popup near the original annotation.
 * <p>
 * When a suggestion is chosen this is reported via  suggestionSelectionListener.suggestionSelected(...).
 * When the user clicks in a non-suggestion area, this is reported via ... .suggestionIgnored(...).
 * When the window is closed, this is reported via ... .suggestionClosed(...).
 * <p>
 * The usual listener is {@link InterlinearMediator}.
 * <pre>
 * +---------------------------------------------------------------+
 * |                                                          close|
 * | +-JScrollPane-----------------------------------------------+ |
 * | |SuggestionSetSelector                                      | |
 * | |                                                           | |
 * | +-----------------------------------------------------------+ |
 * |                                                              R|
 * +---------------------------------------------------------------+
 * </pre>
 * The close "box" and "Resize corner" are handled manually here,
 * as is the window dragging.
 * <p>
 * There is some keyboard control for the selector:
 * <ul>
 * <li>The DIGIT keys 1-9 and 0 select suggestion number 1-10.
 * <li>The SHIFTTED DIGIT keys select suggestions 11-20.
 * <li>The SPACE and ENTER keys select the first suggestion.
 * <li>The TAB, BACKSPACE and DELETE keys ignore this suggestion and move to
 *     the next one (equivalent to clicking outside the suggestions).
 *  <li>The ESCAPE and W keys closes the window and stops the suggestions
 *     (equivalent to clicking the small "close" label in the corner).
 * <li>The ARROW keys scroll the scroll area.
 * <li>The PAGE UP and DOWN keys page through the scroll area.
 * </ul>
 * @author Han Sloetjes
 * @author Olaf Seibert
 */
@SuppressWarnings("serial")
public class SuggestionWindow extends JFrame /*JWindow*/ implements MouseListener, 
MouseMotionListener, ComponentListener, KeyListener,
MouseAndMotionListener, ClientLogger {
	private static final int MAX_INITIAL_WIDTH = 1538;// not sure what this is based on?
	private static final int MAX_INITIAL_HEIGHT = 768;
	private JScrollPane selectorScrollPane; 
	private SuggestionSetSelector selector;
	private SuggestionSelectionListener suggestionSelectionListener;
	private JLabel closeLabel;
	
    private Point dragStartPoint;
    private Point dragEndPoint;
    private final static int DRAG_WINDOW = 0;
    private final static int DRAG_TOP = 1;
    private final static int DRAG_BOTTOM = 2;
    private final static int DRAG_LEFT = 4;
    private final static int DRAG_RIGHT = 8;
    private int dragCorner = DRAG_WINDOW;
    private Insets windowMargins;
    private final int MARGIN_TO_SCROLLBAR = 5;
    private int scrollBarWidth;
    
    // State
	
	/**
	 * Constructor with a frame as the owner.
	 * 
	 * @param frame the owner
	 * @param selector the suggestion set selector component
	 */
	public SuggestionWindow(Frame frame, SuggestionSetSelector selector) {
		//super(frame);
		super();
		this.selector = selector;
		initComponents();
	}

	/**
	 * Constructor with a window as the owner.
	 * 
	 * @param window the owner
	 * @param selector the suggestion set selector component
	 */
	public SuggestionWindow(Window window, SuggestionSetSelector selector) {
		//super(window);
		super();
		this.selector = selector;
		initComponents();
	}
	
	/**
	 * Initializes the ui components, adds the listeners.
	 */
	private void initComponents() {
		// Make JFrame a window without title bar, but hopefully with resizer in the south-east corner.
		setUndecorated(true);
		//setResizable(true);
		
		//setBackground(Color.MAGENTA); // TEST
		selectorScrollPane = new JScrollPane(selector);
		getContentPane().setLayout(null);
		getContentPane().add(selectorScrollPane);
		closeLabel = new JLabel("Close");
		closeLabel.setFont(closeLabel.getFont().deriveFont(8f));
		getContentPane().add(closeLabel);
		
		windowMargins = new Insets(12, 10, 10, 10);
		
		if (false) {
			RepaintManager repaintManager = RepaintManager.currentManager(getRootPane());
			repaintManager.setDoubleBufferingEnabled(false);
			getRootPane().setDebugGraphicsOptions(DebugGraphics.BUFFERED_OPTION | DebugGraphics.FLASH_OPTION
			        /*| DebugGraphics.LOG_OPTION*/);

		}
		if (false) {
			getRootPane().setOpaque(false);	 // TESTING: where is all the flashing coming from??
			getLayeredPane().setOpaque(false);	 // TESTING: where is all the flashing coming from??
			if (getContentPane() instanceof JComponent)
			 {
				((JComponent)getContentPane()).setOpaque(false);	 // TESTING: where is all the flashing coming from??
			}
			selectorScrollPane.setOpaque(false);	 // TESTING: where is all the flashing coming from??
			selectorScrollPane.getViewport().setOpaque(false);	 // TESTING: where is all the flashing coming from??
			selector.setOpaque(false);	 // TESTING: where is all the flashing coming from??
		}			
		if (false) {
			RepaintManager repaintManager = RepaintManager.currentManager(selectorScrollPane);
			repaintManager.setDoubleBufferingEnabled(false);
			getRootPane().setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
			getLayeredPane().setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
			if (getContentPane() instanceof JComponent)
			 {
				((JComponent)getContentPane()).setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
			}
			selectorScrollPane.setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
			selectorScrollPane.getViewport().setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
			selector.setDoubleBuffered(false);	 // TESTING: where is all the flashing coming from??
		}
		if (false) {
			RepaintManager repaintManager = RepaintManager.currentManager(selectorScrollPane);
			repaintManager.setDoubleBufferingEnabled(true);
			getRootPane().setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
			getLayeredPane().setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
			if (getContentPane() instanceof JComponent)
			 {
				((JComponent)getContentPane()).setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
			}
			selectorScrollPane.setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
			selectorScrollPane.getViewport().setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
			selector.setDoubleBuffered(true);	 // TESTING: where is all the flashing coming from??
		}
		
		Object sbWidth = UIManager.getDefaults().get("ScrollBar.width");
        if (sbWidth instanceof Integer) {
            scrollBarWidth = (Integer) sbWidth;
        }
		selectorScrollPane.addMouseListener(this);
		selectorScrollPane.addMouseMotionListener(this);
		closeLabel.addMouseListener(this);
		
		getContentPane().addMouseListener(this);
		getContentPane().addMouseMotionListener(this);
		addComponentListener(this);

		setFocusable(true);	// we want to see keys
		setFocusTraversalKeysEnabled(false); // we want to see the TAB key
		selectorScrollPane.requestFocusInWindow();
		addKeyListener(this);
		
		positionComponents();
	}
	
	/**
	 * Repositions the components of this window.
	 */
	private void positionComponents() {
		int w = getContentPane().getWidth();
		int h = getContentPane().getHeight();
		
		if (w <= 0 || h <= 0) {
			return;
		}
		
		Dimension closeDim = closeLabel.getPreferredSize();
		int cx = w - closeDim.width - windowMargins.right;
		if (cx > windowMargins.left) {
			if (!closeLabel.isVisible()) {
				closeLabel.setVisible(true);
			}
			closeLabel.setBounds(cx, 0, closeDim.width, closeDim.height);
		} else {
			//closeLabel.setVisible(false);
		}
		if (w > windowMargins.left + windowMargins.right && h > windowMargins.top + windowMargins.bottom) {
			if (!selectorScrollPane.isVisible()) {
				selectorScrollPane.setVisible(true);
			}
			selectorScrollPane.setBounds(windowMargins.left, windowMargins.top, w - windowMargins.left - windowMargins.right, 
					h - windowMargins.top - windowMargins.bottom);
			selector.revalidate(); // forces a relayout which may not happen if there are scrollbars
			// and the scrollpane thinks it makes no difference.
			//selectorScrollPane.revalidate(); // seems to help to position the scrollbars
		} else {
			//suggestionScrollPane.setVisible(false);
		}
	}

	/**
	 * Removes the listeners from this window.
	 */
	private void removeListeners() {
		getContentPane().removeMouseListener(this);
		getContentPane().removeMouseMotionListener(this);
		selectorScrollPane.removeMouseListener(this);
		selectorScrollPane.removeMouseMotionListener(this);
		closeLabel.removeMouseListener(this);
		selector.removeWorkAroundToolTipBug(this);
		removeKeyListener(this);
		suggestionSelectionListener = null;
	}
	
	/**
	 * Removes listeners and closes the window.
	 */
	private void closeSelectionWindow() {
		removeListeners();
		setVisible(false);
		dispose();
	}
	
	/**
	 * Sets the listener to be notified of a suggestion selection event.
	 * Currently there can only be one such listener.
	 * 
	 * @param listener the new suggestion selection listener
	 */
	public void setSuggestionSelectionListener(SuggestionSelectionListener listener) {
		suggestionSelectionListener = listener;
	}

	@Override
	public Dimension getPreferredSize() {
		if (selector != null) { // always
			Dimension d = selector.getPreferredSize();
			Dimension maxDim = getMaximumSize();
			if (maxDim == null || maxDim.width == Integer.MAX_VALUE || 
					maxDim.height == Integer.MAX_VALUE) {
				// no maximum size set by the parent component, the creator of this window
				if (d.width > MAX_INITIAL_WIDTH) {
					d.width = MAX_INITIAL_WIDTH;
				}
				if (d.height > MAX_INITIAL_HEIGHT) {
					d.height = MAX_INITIAL_HEIGHT;
				}
			} else {// a maximum area has been set
				if (d.height > maxDim.height + windowMargins.top + windowMargins.bottom + scrollBarWidth) {
					int origPrefHeight = d.height;
					d.height = maxDim.height - windowMargins.top - windowMargins.bottom - scrollBarWidth;
					// if the height is (much) more than the available height check if the width can be increased
					// it is difficult to estimate the ideal size (this would require detailed information from the selector).
					// roughly estimate how many units do not fit
					int numMissing = (int) Math.ceil((origPrefHeight - d.height) / (double)selector.getUnitHeight());
					int numInCurrentViewHeight = (int) Math.ceil(d.height / (double)selector.getUnitHeight());
					int numExtraColumns = (int) Math.ceil(numMissing / (double) numInCurrentViewHeight);
					// assume the average width of a unit is twice the height
					int extraWidth = numExtraColumns * 2 * selector.getUnitHeight();
					d.width += extraWidth;
				}
				if (d.width > maxDim.width + windowMargins.left + windowMargins.right + scrollBarWidth + MARGIN_TO_SCROLLBAR) {
					d.width = maxDim.width - windowMargins.left - windowMargins.right - scrollBarWidth - MARGIN_TO_SCROLLBAR;
					// the height could probably be adjusted (decreased) now
				}
			}
			return new Dimension(d.width  + windowMargins.left + windowMargins.right  + scrollBarWidth + MARGIN_TO_SCROLLBAR,
					             d.height + windowMargins.top  + windowMargins.bottom + scrollBarWidth);
		} else {
			return super.getPreferredSize();
		}
	}
	
	// ######## mouse motion listener for dragging/resizing the window
	
	@Override // MouseMotionListener
	public void mouseDragged(MouseEvent e) {
		dragEndPoint = e.getPoint();
		int dx = dragEndPoint.x - dragStartPoint.x;
		int dy = dragEndPoint.y - dragStartPoint.y;
		Rectangle bounds = getBounds(); // a copy
		
		// Use relative movements, because the coordinates could be in different spaces.
		if (dragCorner == DRAG_WINDOW) {
			bounds.x += dx;
			bounds.y += dy;
			// Don't set dragStartPoint to dragEndPoint, because we have effectively moved
			// the window under the cursor back to dragStartPoint.
		} else {
			if ((dragCorner & DRAG_LEFT) != 0) {
				bounds.x += dx;
				bounds.width -= dx;
			}
			if ((dragCorner & DRAG_RIGHT) != 0) {
				bounds.width += dx;
				dragStartPoint.x = dragEndPoint.x;
			}
			if ((dragCorner & DRAG_TOP) != 0) {
				bounds.y += dy;
				bounds.height -= dy;
			}
			if ((dragCorner & DRAG_BOTTOM) != 0) {
				bounds.height += dy;
				dragStartPoint.y = dragEndPoint.y;
			}
		}
		setBounds(bounds);
	}

	@Override // MouseMotionListener
	public void mouseMoved(MouseEvent e) {
	}

	// ####### mouse listener
	@Override // MouseListener
	public void mouseClicked(MouseEvent e) {
		final Component source = (JComponent) e.getSource();
		
		if (source == closeLabel) {
			if (suggestionSelectionListener != null) {
				SuggestionSelectionEvent event = new SuggestionSelectionEvent(selector.getModel(), -1);
				suggestionSelectionListener.suggestionClosed(event);
			}
			closeSelectionWindow();
		} else if (source == selectorScrollPane) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mouseClicked: Scrollpane %s",
						String.valueOf(e.getPoint())));
			}
			if (e.isShiftDown() || e.isAltDown()) {
				increaseChoiceOfSuggestions();
			} else {
				// There isn't really any need to check.
				// If the source is the scroll pane but but not a SuggestionComponent,
				// the user clicked outside of them and selectedIndex will be -1.
				int xp = e.getPoint().x + selectorScrollPane.getHorizontalScrollBar().getValue();
				int xy = e.getPoint().y + selectorScrollPane.getVerticalScrollBar().getValue();
				
				int selectedIndex = selector.getSuggestionIndexAtPoint(xp, xy);
				if (LOG.isLoggable(Level.FINER)) {
					LOG.finer(String.format("Selected index: " + selectedIndex));
				}
				if (suggestionSelectionListener != null) {
					SuggestionSelectionEvent event = new SuggestionSelectionEvent(selector.getModel(), selectedIndex);
					if (selectedIndex > -1) {
						suggestionSelectionListener.suggestionSelected(event);
					} else {
						suggestionSelectionListener.suggestionIgnored(event);
					}
				}
				closeSelectionWindow();
			}
		} else if (source instanceof SuggestionComponent) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mouseClicked: SuggestionComponent %s",
						String.valueOf(e.getPoint())));
			}
			
			if (e.isShiftDown() || e.isAltDown()) {
				reduceChoiceOfSuggestions((SuggestionComponent)source);
			} else {
				if (suggestionSelectionListener != null) {
					int selectedIndex = ((SuggestionComponent)source).getIndex();
					SuggestionSelectionEvent event = new SuggestionSelectionEvent(selector.getModel(), selectedIndex);
					suggestionSelectionListener.suggestionSelected(event);
				}
				closeSelectionWindow();
			}
		} else if (source == getContentPane()) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mouseClicked: contentPane %s",
						String.valueOf(e.getPoint())));
			}
		} else {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mouseClicked: other source %s",
						String.valueOf(e)));
			}
		}
	}

	/**
	 * The user has made a part-choice.
	 * <p>
	 * Keep all the suggestions that have the same (hash) value for the current
	 * fragment. Discard all others. This is simply done by making them invisible.
	 * 
	 * @param selectedIndex
	 * @param selected
	 */
	private void reduceChoiceOfSuggestions(SuggestionComponent selected) {
		int fragNr = selected.getFragNr();
		int hash = selected.getHashOfFrag(fragNr);
		int nextFragNr = fragNr >= 0 ? fragNr + 1
				                     : fragNr - 1;
		
		Component[] components = selector.getComponents();
		int nVisible = 0;
		
		for (int i = components.length - 1; i >= 0; i--) {
			Component c = components[i];
			if (c.isVisible() && c instanceof SuggestionComponent) {
				SuggestionComponent sc = (SuggestionComponent)c;
				if (hash == sc.getHashOfFrag(fragNr)) {
					sc.setFragNr(nextFragNr);
					nVisible += 1;
				} else {
					c.setVisible(false);
				}
			}
		}
		
		if (nVisible > 1) {
		} else if (nVisible == 1) {
			int selectedIndex = selected.getIndex();
			SuggestionSelectionEvent event = new SuggestionSelectionEvent(selector.getModel(), selectedIndex);
			suggestionSelectionListener.suggestionSelected(event);
			closeSelectionWindow();
		} else { // nVisible == 0
			// This can't happen. At least component #selectedIndex should still be there.
		}
		
		selector.revalidate();
		selectorScrollPane.revalidate();
		selector.revalidate();
		invalidate();
		repaint();
	}

	/**
	 * Increase the choice of suggestions by making them all visible again.
	 */
	private void increaseChoiceOfSuggestions() {
		Component[] components = selector.getComponents();
		
		for (int i = components.length - 1; i >= 0; i--) {
			Component c = components[i];
			if (c instanceof SuggestionComponent) {
				c.setVisible(true);
			}
		}
	}

	@Override // MouseListener
	public void mouseEntered(MouseEvent e) {
	}

	@Override // MouseListener
	public void mouseExited(MouseEvent e) {
	}

	@Override // MouseListener
	public void mousePressed(MouseEvent e) {
		final Container contentPane = getContentPane();
		// We get notifications from the content pane and from the scroll pane within it,
		// which may have slightly offset coordinates.
		// The dragging coordinates will be in the same space, and they are used relatively,
		// so there is no need to translate them.
		dragStartPoint = e.getPoint();
		final Object source = e.getSource();
		
		// However the resize corner is outside the scroll pane anyway.
		if (source == contentPane) {
			final int width = contentPane.getWidth();
			final int height = contentPane.getHeight();
			
			dragCorner = DRAG_WINDOW;
			if (dragStartPoint.x < windowMargins.left) {
				dragCorner |= DRAG_LEFT;
			} else if (dragStartPoint.x > width - windowMargins.right) {
				dragCorner |= DRAG_RIGHT;
			}
			if (dragStartPoint.y < windowMargins.top) {
				dragCorner |= DRAG_TOP;
			} else if (dragStartPoint.y > height - windowMargins.bottom) {
				dragCorner |= DRAG_BOTTOM;
			}
		} else if (e.getSource() == selectorScrollPane) {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mousePressed: Scrollpane %s", e.getPoint().toString()));
			}
		} else {
			if (LOG.isLoggable(Level.FINER)) {
				LOG.finer(String.format("SuggestionWindow.mousePressed: other source %s", e.toString()));
			}
		}
	}

	@Override // MouseListener
	public void mouseReleased(MouseEvent e) {
		dragCorner = DRAG_WINDOW;
	}

	@Override // ComponentListener
	public void componentHidden(ComponentEvent e) {
	}

	@Override // ComponentListener
	public void componentMoved(ComponentEvent e) {
	}

	@Override // ComponentListener
	public void componentResized(ComponentEvent e) {
		positionComponents();
	}

	@Override // ComponentListener
	public void componentShown(ComponentEvent e) {
		positionComponents(); // a bit unneeded since the setSize() will cause componentResized() to be called
		selector.init(); // causes it to know its size and contents
		selector.workAroundTooltipBug(this);
		
		final int unitHeight = selector.getUnitHeight();
		selectorScrollPane.getVerticalScrollBar().setUnitIncrement(unitHeight);
		selectorScrollPane.getHorizontalScrollBar().setUnitIncrement(unitHeight);
		selectorScrollPane.revalidate();
		this.setSize(getPreferredSize());
	}

	/**
	 * Use a KeyListener instead of InputMap and ActionMap because there is no
	 * access to the actually typed key (especially shift-0...9), without making
	 * a separate Action for each key.
	 */
	@Override // KeyListener
	public void keyTyped(KeyEvent e) {
		if (LOG.isLoggable(Level.FINER)) {
			LOG.finer(String.format("SuggestionWindow.keyTyped %s", String.valueOf(e)));
		}
	}

	@Override // KeyListener
	public void keyPressed(KeyEvent e) {
		if (LOG.isLoggable(Level.FINER)) {
			LOG.finer(String.format("SuggestionWindow.keyPressed %s", String.valueOf(e)));
		}
		
		if (suggestionSelectionListener != null) {
			final IGTSuggestionViewerModel model = selector.getModel();

			int selectedIndex = 1; // default; find out which key below.

			SuggestionSelectionEvent event;
			JScrollBar sb;
			int code = e.getKeyCode();
			
			switch (code) {
			case KeyEvent.VK_SPACE:
			case KeyEvent.VK_ENTER:
				code = KeyEvent.VK_1;
				// fall through; pretend space was 1.
			case KeyEvent.VK_0:	case KeyEvent.VK_1:	case KeyEvent.VK_2:
			case KeyEvent.VK_3:	case KeyEvent.VK_4:	case KeyEvent.VK_5:
			case KeyEvent.VK_6:	case KeyEvent.VK_7:	case KeyEvent.VK_8:
			case KeyEvent.VK_9:
				selectedIndex = code - KeyEvent.VK_0;
				if (selectedIndex == 0) {
					selectedIndex = 10;
				}
				if (e.isShiftDown()) {
					selectedIndex += 10;
				}
				
				if (selectedIndex <= model.getRowCount()) {
					event = new SuggestionSelectionEvent(model, selectedIndex - 1);
					if (LOG.isLoggable(Level.FINER)) {
						LOG.finer(String.format("suggestionSelected %d\n", selectedIndex - 1));
					}
					suggestionSelectionListener.suggestionSelected(event);
					closeSelectionWindow();
				}
				break;
				
			case KeyEvent.VK_TAB: // only gets reported when setFocusTraversalKeysEnabled(false)
			case KeyEvent.VK_BACK_SPACE:
			case KeyEvent.VK_DELETE:
				event = new SuggestionSelectionEvent(model, -1);
				LOG.finer("suggestionIgnored");
				suggestionSelectionListener.suggestionIgnored(event);
				closeSelectionWindow();
				break;
				
			case KeyEvent.VK_ESCAPE:
			case KeyEvent.VK_W: // also matches Command-W
				event = new SuggestionSelectionEvent(model, -1);
				LOG.finer("suggestionClosed");
				suggestionSelectionListener.suggestionClosed(event);
				closeSelectionWindow();
				break;

				// This should not be needed for keyboard control of the
				// scroll area, but I didn't get it to work otherwise.
			case KeyEvent.VK_PAGE_UP:
				sb = selectorScrollPane.getVerticalScrollBar();
				sb.setValue(sb.getValue() - sb.getBlockIncrement(-1));
				break;
			
			case KeyEvent.VK_PAGE_DOWN:
				sb = selectorScrollPane.getVerticalScrollBar();
				sb.setValue(sb.getValue() + sb.getBlockIncrement(1));
				break;

			case KeyEvent.VK_UP:
				sb = selectorScrollPane.getVerticalScrollBar();
				sb.setValue(sb.getValue() - sb.getUnitIncrement(-1));
				break;
			
			case KeyEvent.VK_DOWN:
				sb = selectorScrollPane.getVerticalScrollBar();
				sb.setValue(sb.getValue() + sb.getUnitIncrement(1));
				break;

			case KeyEvent.VK_LEFT:
				sb = selectorScrollPane.getHorizontalScrollBar();
				sb.setValue(sb.getValue() - sb.getUnitIncrement(-1));
				break;
			
			case KeyEvent.VK_RIGHT:
				sb = selectorScrollPane.getHorizontalScrollBar();
				sb.setValue(sb.getValue() + sb.getUnitIncrement(1));
				break;
			}
		}
	}

	@Override // KeyListener
	public void keyReleased(KeyEvent e) {
		if (LOG.isLoggable(Level.FINER)) {
			LOG.finer(String.format("SuggestionWindow.keyReleased %s", String.valueOf(e)));
		}
	}

	/** for setting a breakpoint */
//	@Override
//	public void update(Graphics g) {
//		super.update(g);
//	}
}
