package mpi.eudico.client.annotator.interlinear.edit;

import static mpi.eudico.client.annotator.util.ClientLogger.LOG;

import java.awt.Component;
import java.io.File;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Queue;
import java.util.Stack;
import java.util.logging.Level;

import mpi.eudico.client.annotator.Constants;
import mpi.eudico.client.annotator.commands.AnnotationsFromSuggestionSetCommand;
import mpi.eudico.client.annotator.commands.Command;
import mpi.eudico.client.annotator.commands.ELANCommandFactory;
import mpi.eudico.client.annotator.interlinear.edit.config.AnalyzerConfig;
import mpi.eudico.client.annotator.interlinear.edit.config.AnalyzerTypeConfig;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSelectionListener;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSetEvent;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSetListener;
import mpi.eudico.client.annotator.interlinear.edit.event.SuggestionSetProvider;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionModel;
import mpi.eudico.client.annotator.interlinear.edit.model.IGTSuggestionViewerModel;
import mpi.eudico.client.annotator.layout.InterlinearizationManager;
import mpi.eudico.server.corpora.clom.Tier;
import mpi.eudico.server.corpora.clom.Transcription;
import mpi.eudico.server.corpora.clomimpl.abstr.AbstractAnnotation;
import mpi.eudico.server.corpora.clomimpl.abstr.TierImpl;
import mpi.eudico.server.corpora.clomimpl.type.Constraint;
import mpi.eudico.server.corpora.clomimpl.type.LinguisticType;
import mpi.eudico.server.corpora.lexicon.LexiconQueryBundle2;
import nl.mpi.lexan.analyzers.LexiconTextAnalyzer;
import nl.mpi.lexan.analyzers.TextAnalyzer;
import nl.mpi.lexan.analyzers.TextAnalyzerContext;
import nl.mpi.lexan.analyzers.helpers.Information;
import nl.mpi.lexan.analyzers.helpers.LexanSuggestionSelectionListener;
import nl.mpi.lexan.analyzers.helpers.Position;
import nl.mpi.lexan.analyzers.helpers.PositionLexicon;
import nl.mpi.lexan.analyzers.helpers.SourceTargetConfiguration;
import nl.mpi.lexan.analyzers.helpers.Suggestion;
import nl.mpi.lexan.analyzers.helpers.SuggestionSet;
import nl.mpi.lexan.analyzers.helpers.prompt.Prompt;

/**
 * An Analyzer Host Context, the context for TextAnalyzers to communicate with.
 * 
 * This is to some extend a combination of old Mediator and Controller functionality.
 */
public class TextAnalyzerHostContext 
	implements TextAnalyzerContext, SuggestionSelectionListener, 
				SuggestionSetProvider {
	private File analyzerCacheFolder = new File(Constants.ELAN_DATA_DIR, Constants.ANALYZER_CACHE_FOLDER_NAME);
	private InterlinearizationManager manager;
	private Transcription transcription;
	private List<SuggestionSetListener> suggestionListeners;
	// administration of loaded analyzers
	private List<TextAnalyzer> loadedAnalyzers;// analyzers for actual use with source and target set
	private Map<Information, Component> info2configComp;
	
	// temp
	private List<AnalyzerConfig> configs;
	private boolean recursive = false;

    /**
     * A queue of suggestions that still need to be shown to the user in a window. 
     * Each element was an argument to {@link #newAnnotations(List)}.
     */
    private Queue<List<SuggestionSet>> deferredNewAnnotations;
	private boolean sugWindowOpened = false; // Could also be called 'deferNewAnnotations'
	
	public TextAnalyzerHostContext(InterlinearizationManager manager) {
		super();
		this.manager = manager;
		this.transcription =  manager.getTranscription();
		loadedAnalyzers = new ArrayList<TextAnalyzer>();
		info2configComp = new HashMap<Information, Component>();
		configs = new ArrayList<AnalyzerConfig>(6);
		suggestionListeners = new ArrayList<SuggestionSetListener>(4);
		ensureCacheFolder();
	}
	
	/**
	 * Checks if the cache folder exists and creates it if not.
	 */
	public void ensureCacheFolder() {
		if (analyzerCacheFolder == null) {
			analyzerCacheFolder = new File(Constants.ELAN_DATA_DIR, 
					Constants.ANALYZER_CACHE_FOLDER_NAME);
		}
		
		try {
			if (!analyzerCacheFolder.exists()) {
				try {
					analyzerCacheFolder.mkdir();
				} catch (Throwable t) {// any
					if (LOG.isLoggable(Level.WARNING)) {
						LOG.warning(String.format("Cannot create the cache folder for analyzers: %s", 
								t.getMessage()));
					}
					return;
				}
			}
		} catch (Throwable thr) {
			if (LOG.isLoggable(Level.WARNING)) {
				LOG.warning(String.format("Cannot check the existence of the analyzers cache folder: %s", 
						thr.getMessage()));
			}
		}
	}
	
	public Transcription getTranscription() {
		return transcription;
	}
	
	/**
	 * Adds a configuration to the list of tier-analyzer configurations.
	 * <p>
	 * This gets called from the preferences interpreting code.
	 * This also means that it gets called again when preferences change.<br/>
	 * This is nasty.<br/>
	 * To counter this, perform a check for duplications (and disallow them).
	 *  
	 * @param config
	 */
	public void addConfig(AnalyzerConfig config) {
		if (!configs.contains(config)) {
			configs.add(config);
			
			List<SourceTargetConfiguration> sourceTargetConfList = convertToSTConfList(config);

			TextAnalyzer ta = TextAnalyzerRegistry.getInstance().getAnalyzerForDoc(
					transcription.getURN().toString(), config.getAnnotId());
			boolean success = false;
			
			if (ta != null) {
				if (loadedAnalyzers.contains(ta)) {
					// Add to the already loaded configuration.
					// (Alternatively: do ta.unload() and ta.load() with ALL configs on `configs'.)
					success = ta.partLoad(sourceTargetConfList);
				} else {
					// temp call loadAnalyzer for now. Has to be decided how best to init / load analyzers
					// loadAnalyzer could return a boolean instead
					success = (ta == loadAnalyzer(config.getAnnotId(), sourceTargetConfList));
				}
			}

			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine(String.format("Initialized annotation analyzer '%s': %s\n", 
						config.getAnnotId().getName(), success));
			}

			// notify the editor
			manager.getInterEditor().configsChanged();
		}
	}
	
	/**
	 * Creates a Position or a PositionLexicon object for the specified tier.
	 * 
	 * @param t the tier to create a Position object for
	 * @return a {@link Position} or {@link PositionLexicon} object
	 */
	private Position tierToPosition (Tier tier) {
		if (tier == null) {
			return null;
		}
		Position pos = null;
		
		LinguisticType sourceType = tier.getLinguisticType();
		// the query bundle, if not null, contains the name of the lexicon and of the entry field 
		LexiconQueryBundle2 lexBundle = sourceType.getLexiconQueryBundle();
		if (lexBundle != null) {
			// should we check if LexiconLink and LexiconIdentification are not null?
			String lexiconName = lexBundle.getLink().getLexId().getName();
			String field = lexBundle.getFldId().getName();
			if (lexiconName != null && field != null) {
				pos = new PositionLexicon(tier.getName(), lexiconName, field);
			} else {
				pos = new Position(tier.getName());
			}
		} else {
			pos = new Position(tier.getName());
		}
		
		return pos;
	}
	
	/**
	 * Converts a tier based configuration (one source tier, one or more target tiers)
	 * to a LEXAN API source-target configuration object.
	 * 
	 * @param ac the configurations to convert
	 * @return a SourceTargetConfiguration object or null
	 */
	private SourceTargetConfiguration convertToSTConfiguration(AnalyzerConfig ac) {
		if (ac == null || ac instanceof AnalyzerTypeConfig) {
			return null;
		}
		// one source, one or more destination objects
		Tier sourceTier = transcription.getTierWithId(ac.getSource());
		if (sourceTier == null) {
			// add logging
			return null;
		}
		Position sourcePos = tierToPosition(sourceTier);
		if (sourcePos == null) {
			return null;
		}
		List<Position> targetPosList = new ArrayList<Position>(2);
		for(String targetName : ac.getDest()) {
			Tier targetTier = transcription.getTierWithId(targetName);
			Position targetPos = tierToPosition(targetTier);
			if (targetPos != null) {
				targetPosList.add(targetPos);
			}
		}
		// must be at least one target tier, at the moment
		if (targetPosList.isEmpty()) {
			return null;
		}
		
		SourceTargetConfiguration stConfig = new SourceTargetConfiguration(sourcePos, targetPosList);
		stConfig.setDocumentId(transcription.getURN().toString());
		
		return stConfig;
	}
	
	/**
	 * Converts an analyzer configuration object to a list of LEXAN API source-target 
	 * configuration objects. 
	 * 
	 * @param ac a tier-based or a type-based configuration. In the latter case there 
	 * might be more than one source tiers and the returned list's size might be > 1.
	 * 
	 * @return a list of source-target configuration objects (or null if the input is null)
	 */
	private List<SourceTargetConfiguration> convertToSTConfList(AnalyzerConfig ac) {
		if (ac == null) {
			return null;
		}
		List<SourceTargetConfiguration> sourceTargetList = new ArrayList<SourceTargetConfiguration>();
		List<AnalyzerConfig> acTierConfList = ac.getTierConfigurations();
		
		for (AnalyzerConfig anaCon : acTierConfList) {
			SourceTargetConfiguration stc = convertToSTConfiguration(anaCon);
			if (stc != null) {
				sourceTargetList.add(stc);
			}
		}
		
		return sourceTargetList;
	}
	
	/**
	 * Removes a configuration from an analyzer. This can currently not be used in case of 
	 * removal of a tier.
	 * 
	 * @param config the configuration to remove
	 */
	public void removeConfig(AnalyzerConfig config) {
		if (configs.remove(config)) {
			
			List<SourceTargetConfiguration> sourceTargetConfList = convertToSTConfList(config);
			if (sourceTargetConfList.isEmpty()) {
				// log
				return;
			}

			unloadAnalyzer(config.getAnnotId(), sourceTargetConfList);
		}
	}
	
    /**
	 * Gives access to the list of configurations
	 * @return
	 */
	public List<AnalyzerConfig> getConfigurations() {
		return configs;
	}
	
	/**
	 * The list of available text analyzers as discovered in the extensions location(s).
	 * 
	 * @return a list of analyzers information objects 
	 */
	public List<Information> listTextAnalyzersInfo() {
		return TextAnalyzerRegistry.getInstance().getAnalyzersInfo();
	}
	
	/**
	 * Load an analyzer (used from addConfig(AnalyzerConfig)).
	 * 
	 * @param info descriptor for the analyzer
	 * @param stConfigs the source-target configurations which include lexicon information if applicable
	 * 
	 * @return the analyzer instance if loaded successfully
	 */
    private TextAnalyzer loadAnalyzer(Information info, List<SourceTargetConfiguration> stConfigs) {  	
        if (stConfigs == null || stConfigs.isEmpty()) {
        	// log error throw exception?
        	return null;
        }

        TextAnalyzer ta = TextAnalyzerRegistry.getInstance().getAnalyzerForDoc(
        		transcription.getURN().toString(), info);
        
        if (ta != null) {
        	prepareAnalyzer(ta);
        	if (ta.load(stConfigs)) {	
        		loadedAnalyzers.add(ta);
        		
        		return ta;
        	}       	
        }
    	return null;
    }
    
    
    /**
     * Unloads an analyzer, which means it is disconnected from the source(s) and target(s).
     * @param info the information object of the analyzer
     * @param source the source tiers
     * @param target the target tiers
     */
    public void unloadAnalyzer(Information info, List<SourceTargetConfiguration> stConfigs) {
        TextAnalyzer toUnLoad = null;
        for (TextAnalyzer a : loadedAnalyzers) {
        	// the Information of the analyzer is a different instance, so cannot test with ==
            if (info.equals(a.getInformation())) {
            	toUnLoad = a;
            	break;
            }
        }
        
        if (toUnLoad == null) {
        	return;
        }
        // Check if we want to completely unload or partially unload??
        
        if (toUnLoad.partUnload(stConfigs) == 0) {
            toUnLoad.unload();
            loadedAnalyzers.remove(toUnLoad);//??

            TextAnalyzerRegistry.getInstance().removeAnalyzerForDoc(
            		transcription.getURN().toString(), info);
            return;
        }
        // Just partially unloaded
    }
	
    @Override
    public Component getConfigurationComponent(Information info, boolean loadIt) {
    	if (info2configComp.containsKey(info)) { // also succeeds when value == null
    		return info2configComp.get(info);
    	}
    	
    	for (TextAnalyzer a: loadedAnalyzers) {
            if (info == a.getInformation()) { 
            	// Some analyzers need that setCacheFolder() has been called on them.
            	// Similarly for setLexiconContext().
            	Component c = a.getConfigurationComponent();
            	info2configComp.put(info, c);
            	return c;
            }
        }

    	if (loadIt) {
	    	// load the analyzer and get the component? Not sure when this can happen
	    	TextAnalyzer ta = TextAnalyzerRegistry.getInstance().getAnalyzerForDoc(
	    			transcription.getURN().toString(), info);
			if (ta != null) {
				// Add the analyzer to the list of loaded analyzers??
				// Maybe if we can load() it without a configuration, which isn't the case (yet?).
				prepareAnalyzer(ta);
				Component c = ta.getConfigurationComponent();
				if (c != null) {
					info2configComp.put(info, c);
				}
				
				return c;
			}
    	}
    	
    	return null;
    }

	/**
	 * Prepare an analyzer for use by giving it some context information it needs for its work,
	 * and which the configuration component needs.
	 * <p>
	 * Implementation note: used both for loadAnalyzer() and getConfigurationComponent().
	 * For the latter, the analyzer itself is no longer needed, and should not be prevented from
	 * being garbage-collected by what we do here.
	 * On the other hand it may be a cached instance, so it should not explicitly be
	 * deconfigured.
	 */
	private void prepareAnalyzer(TextAnalyzer ta) {
    	ta.setAnalyzerContext(this);
		ta.setCacheFolder(analyzerCacheFolder);
		if (ta instanceof LexiconTextAnalyzer) {
			((LexiconTextAnalyzer) ta).setLexiconContext(manager.getTextAnalyzerLexiconContext());
		}
	}
    
    /**
     * This method must be called before one or more calls to analyze(...).
     * This in turn calls back to our newAnnotation(s) methods.
     * <p>
     * In particular newAnnotations(List<SuggestionSet> sugSets) will enqueue
     * suggestions on the deferredNewAnnotations queue, but the others will
     * simply apply the suggestion immediately.
     */
    public void interLinearizeActionStarting() {
    	if (LOG.isLoggable(Level.FINE)) {
    		LOG.fine("interLinearizeActionStarting: clear queue, close window");
    	}
        if (deferredNewAnnotations != null) {
    		if (LOG.isLoggable(Level.FINE)) {
    			LOG.fine(String.format("interLinearizeActionStarting: queue had %d items", deferredNewAnnotations.size()));
    		}
        }
   		stack = new Stack<Queue<List<SuggestionSet>>>();
    	deferredNewAnnotations = new LinkedList<List<SuggestionSet>>();
    	stack.push(deferredNewAnnotations);
        
        // Here we remove any existing open suggestion windows.
        notifyCancelSuggestionSet();
        sugWindowOpened = false;
    }

	/**
	 * Called when the program is done with generating all suggestions.
	 * <p>
	 * Implementation note:
	 * Typically the GUI has not had a chance to run during all the work of creating
	 * the suggestions. Therefore, although a SuggestionWindow has been created,
	 * the user very likely has not seen it yet! She will be processing all suggestions
	 * at her own pace, as soon as the GUI has a chance to show the window.
	 */
    public void interLinearizeActionCompleted() {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine(String.format("interLinearizeActionCompleted: queue has %d items\n",
				deferredNewAnnotations == null
				                        ? -1
				                        : deferredNewAnnotations.size()));
		}
    }

    private Stack<Queue<List<SuggestionSet>>> stack;
    private int recursionLevel; 
    
    /**
     * Given a suggestion at 'baseRecursionLevel', announce the intent to call
     * analyze() again to generate suggestions from it.
     * <p>
     * In normal cases, a method like this could count its own levels.
     * However here, suggestions may be queued up for later display (but some
     * are processed immediately). For the queued suggestions it is important
     * to know at which level they were, to properly queue suggestions triggered
     * from them...
     * 
     * @param baseRecursionLevel
     * @return the old recursion level, to be passed to interLinearizeActionRecursiveCompleted().
     */
    public int interLinearizeActionRecursiveStarting(int baseRecursionLevel) {
    	int oldLevel = recursionLevel;
    	recursionLevel = baseRecursionLevel + 1;
    	// If we're getting to a new level, push the deferredAnnotations on the stack.
    	if (recursionLevel >= stack.size()) {
	    	deferredNewAnnotations = new LinkedList<List<SuggestionSet>>();
	    	stack.push(deferredNewAnnotations);
	    	// Maintain the invariant that
	    	// deferredNewAnnotations == stack.get(recursionLevel)
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine(String.format("interLinearizeActionRecursiveStarting: level %d, stack now has %d items\n",
					recursionLevel, stack.size()));
			}
    	} else { // recursionLevel < stack.size()
    		// If we're getting back to a pre-existing level, apparently
    		// without the GUI having processed it, simply append to it.
    		deferredNewAnnotations = stack.get(recursionLevel);
    		if (LOG.isLoggable(Level.FINE)) {
    			LOG.fine(String.format("interLinearizeActionRecursiveStarting: level %d, stack still has %d items (not pushed)\n",
					recursionLevel, stack.size()));
    		}
    	}
    	// Not setting sugWindowOpened to false
        // prevents the window being opened from newAnnotations().
        // It will be opened naturally when the parent analysis continues with the next
        // suggestion but the queue has been switched from under its feet.
    	return oldLevel;
    }
    
    /**
     * Also here, the user hasn't seen the window yet...
     * 
     * @param oldLevel The value returned from interLinearizeActionRecursiveStarting().
     */
    public void interLinearizeActionRecursiveCompleted(int oldLevel) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine(String.format("interLinearizeActionRecursiveCompleted: stack %d items, queue has %d items\n",
				stack.size(),
				(deferredNewAnnotations == null
				                         ? -1
				                         : deferredNewAnnotations.size())));
		}
    	recursionLevel = oldLevel;
    	// The below is eventually done in notifyLexanSuggestionClosed():
    	//deferredNewAnnotations = stack.pop();
    	// restart the GUI on the popped queue.
		//sugWindowOpened = true;
    	//showNextSuggestionSet();
    }
    
    /**
     * Call all analyzers with the given position.
     * The ones that are configured to use it as a source, will do their thing.
     * <p>
     * Before calling this for the first time in a sequence, call interLinearizeActionStarting()
     * first.
     * After the sequence, call interLinearizeActionCompleted().
     * This is for the GUI to know which suggestions to show in a window.
     * <p>
     * The analyzers will call back to our methods
     * {@link #newAnnotation(Position, List)},
     * {@link #newAnnotation(Position, Suggestion)},
     * and/or
     * {@link #newAnnotations(List)}.
     *  
     * @param pos
     */
	public void analyze(Position pos) {
        // moved from ControllerImpl:
		// TODO implement get all analyzers with this Position as one if the source positions
		// and call analyze(pos) on those analyzer. As it is, it calls all analyzers and
		// leaves it to their discretion if they do anything (which they would do if they
		// were configured/loaded for this source tier). Which situation it is, remains
		// unknown to us here.
        for (TextAnalyzer a : loadedAnalyzers) {
            a.analyze(pos);
        }
	}
	
	@Override
	public boolean changeAnnotation(Position pos, Suggestion ann) {
		if (pos == null) {
			// warn
			LOG.warning("Change annotation: Position is null");
			return false;
		}
		if (ann == null) {
			// warn
			LOG.warning("Change annotation: no annotation specified");
			return false;
		}
		
		boolean changed = false;
		TierImpl destTier = (TierImpl) transcription.getTierWithId(pos.getTierId());
		
		if (destTier != null) {
			long t = (pos.getBeginTime() + pos.getEndTime()) / 2;
			AbstractAnnotation aa = (AbstractAnnotation) destTier.getAnnotationAtTime(t);
			if (aa != null) {
				String oldText = aa.getValue();
				String newText = ann.getContent();
				if (newText == null) {
					newText = "";
				}
				Command command = ELANCommandFactory.createCommand(transcription,
						ELANCommandFactory.MODIFY_ANNOTATION);
				command.execute(aa, new Object[] { oldText, newText } );
				changed = true;
			} else {
				// warn
				LOG.warning("Change annotation: there is no annotation at the specified position");
			}
		}
		
		return changed;
	}

	/**
	 * Creates one new annotation at the specified position.
	 * In case of a subdivision tier the annotation will be added after the last annotation
	 * in the interval (if it exists).
	 * <p>
	 * This is non-interactive, and so recursive analyze() calls will be performed.
	 */
	@Override
	public boolean newAnnotation(Position pos, Suggestion ann) {
		if (ann == null) {
			// warn
			LOG.warning("New annotation: no new annotation specified");
			return false;
		}

		List<Suggestion> list = new ArrayList<Suggestion>(1);
		list.add(ann);
		return newAnnotation(pos, list);
	}
	
	/**
	 * Creates multiple new annotations at the specified position.
	 * In case of a subdivision tier the annotations will be added after the last annotation
	 * in the interval (if it exists).
	 * <p>
	 * This is non-interactive, and so recursive analyze() calls will be performed.
	 */
	@Override
	public boolean newAnnotation(Position pos, List<Suggestion> annotations) {
		if (pos == null) {
			// warn
			LOG.warning("New annotation: Position is null");
			return false;
		}
		if (annotations == null || annotations.isEmpty()) {
			// warn
			LOG.warning("New annotation: no new annotations specified");
			return false;
		}
		
		boolean created = false;
		TierImpl destTier = (TierImpl) transcription.getTierWithId(pos.getTierId());
		
		if (destTier != null) {
			Constraint cons = destTier.getLinguisticType().getConstraints();
			if (cons == null) {
				// warn, illegal tier type for destination?
				return false;
			}
			
			SuggestionSet sugSet = new SuggestionSet(pos);
			sugSet.addAll(annotations);
			
			// Create annotations via a AnnotationsFromSuggestionSetCommand,
			// so the user can undo/redo it.
			annotationsFromSuggestionSet(sugSet, recursionLevel);
		}
		
		return created;
	}

	/**
	 * Create annotations via a AnnotationsFromSuggestionSetCommand, so the user can undo/redo it.
	 * <p>
	 * Also do recursive analyzing, if enabled.
	 * <p>
	 * Implementation node: when the `subscribing' to tiers actually works (currently
	 * it is unimplemented!), analyzers that
	 * subscribe to their sources will be recursively invoked via that mechanism,
	 * and recursion here will be unneeded.
	 * 
	 * @param sugSet
	 * @param recursionLevel 
	 */
	private void annotationsFromSuggestionSet(SuggestionSet sugSet, int recursionLevel) {
		AnnotationsFromSuggestionSetCommand command = 
				(AnnotationsFromSuggestionSetCommand) ELANCommandFactory.createCommand(
						transcription, ELANCommandFactory.ANNS_FROM_SUGGESTION_SET);
		command.execute(transcription, new Object[]{ sugSet, Boolean.valueOf(recursive) });
		
		if (recursive) {
			List<AbstractAnnotation> recurses = (List<AbstractAnnotation>) command.getCreatedAnnotations();
			
			if (recurses != null) {
				// Start recursive annotations
				int oldLevel = interLinearizeActionRecursiveStarting(recursionLevel);
				
				for (AbstractAnnotation ann : recurses) {
					Position np = new Position(ann.getTier().getName(), 
							ann.getBeginTimeBoundary(), ann.getEndTimeBoundary());

					analyze(np);
				}
				// End recursive annotations
				interLinearizeActionRecursiveCompleted(oldLevel);
			}
		}
	}

    /**
     * If a list of alternative suggestions is produced by one of the extensions, notify listeners.
     * For now this mechanism is used to push the suggestions to the user interface layer.
     * <p>
     * The first call (after {@link #interLinearizeActionStarting()})
     * passes the suggestions on via {@link #notifySuggestionSetDelivered(List)}
     * which likely causes a window to be opened for the user (eventually, when the GUI
     * gets to run). This strategy may hope to
     * <p>
     * On subsequent calls the sugSets are collected for display to the user later. 
     * 
     * @param sugSets the list of suggestion sets. Each set represents a list of possible 
     * values from which the user can pick one. They are put into a window and the user
     * can click.
     * <p>
     * Child suggestions are processed too, but recursive annotation seems not to be done
     * (depending on the SuggestionSetListener). This is because our own 
     * {@link #suggestionSelected(SuggestionSelectionEvent)}
     * does not use our own {@link #newAnnotation(Position, Suggestion) or
     * {@link #newAnnotation(Position, List)}
     * methods.
     * <p>
     * That does make some sense though, since generating suggestion pop-ups from suggestions
     * may be confusing (and current GUI code would not handle it).
     */
	@Override // TextAnalyzerContext
	public boolean newAnnotations(List<SuggestionSet> sugSets) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine(String.format("newAnnotations(): recursionLevel=%d stack has %d items\n",
				recursionLevel, stack.size()));
		}
		if (sugSets != null) {
			if (!sugWindowOpened) {
				// Initial case: show the suggestion right away.
				notifySuggestionSetDelivered(sugSets, recursionLevel);
				sugWindowOpened = true;
				return true;	
            } else {
            	// Later cases: queue them up.
                deferredNewAnnotations.add(sugSets);
			}
		}
		return false;
	}

    /**
     * We received some suggestions and the GUI can act on them.
     * <p>
     * In our practice, the listener is IGTViewer which will open
     * a window and show the suggestions.
     * <p>
     * This will in the end cause our {@link #suggestionSelected(SuggestionSelectionEvent)}
     * method to be called if the user clicks on one of the suggestions,
     * {@link #suggestionIgnored(SuggestionSelectionEvent)}
     * if the user wants to skip that suggestion, 
     * or {@link #suggestionClosed()} if the user dismisses the window.
     * (We are the SuggestionSelectionListener.)
     * 
     * @param sugSets
     * @param recursionLevel 
     */
	private void notifySuggestionSetDelivered(List<SuggestionSet> sugSets, int recursionLevel) {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine(String.format("notifySuggestionSetDelivered: this opens a window; recursionLevel=%d stack has %d items\n",
				recursionLevel, stack.size()));
		}
		SuggestionSetEvent event = new SuggestionSetEvent(this, sugSets, recursionLevel);

		for (SuggestionSetListener ssl : suggestionListeners) {
			ssl.suggestionSetDelivered(event);
			// Eventually our suggestionSelected() method is called
			// if the user clicks on one of the suggestions;
			// suggestionIgnored() if the user clicks outside a suggestion;
			// suggestionClosed() the user dismisses the window.
		}
	}

	/**
	 * This notification is sent when the user just clicked some action
	 * to create new suggestions. Therefore any previously opened GUI windows
	 * for suggestions need to be cancelled.
	 */
	private void notifyCancelSuggestionSet() {
		for (SuggestionSetListener ssl : suggestionListeners) {
			ssl.cancelSuggestionSet();
		}
	}
    
    @Override
	public String prompt(Prompt arg0) {
		// TODO Auto-generated method stub
		return null;
	}

	/**
	 * Returns the annotations from the specified position 
	 * (i.e. from the specified tier strictly within the specified interval).
	 * Use end = -1 for no end time.
	 */
	@Override
	public List<Suggestion> readAnnotation(Position pos) {
		if (pos == null) {
			// warn
			if (LOG.isLoggable(Level.WARNING)) {
				LOG.warning("Read annotations: Position is null");
			}
			return null;
		}
		List<Suggestion> annotations = new ArrayList<Suggestion>();
		
		TierImpl destTier = (TierImpl) transcription.getTierWithId(pos.getTierId());
		
		if (destTier != null) {
			List<AbstractAnnotation> curAnnos = destTier.getAnnotations();
			final long posBeginTime = pos.getBeginTime();
			final long posEndTime = pos.getEndTime() >= 0 ? pos.getEndTime()
					                                      : Long.MAX_VALUE;
			
			for (int i = 0; i < curAnnos.size(); i++) {
				AbstractAnnotation aa = curAnnos.get(i);
				
				final long annoEndTime = aa.getEndTimeBoundary();
				if (annoEndTime < posBeginTime) {
					continue;
				}
				
				final long annoBeginTime = aa.getBeginTimeBoundary();
				if (annoBeginTime > posEndTime) {
					break;
				}
				
				if (annoBeginTime >= posBeginTime && annoEndTime <= posEndTime) {
					Suggestion a = new Suggestion(aa.getValue(),
							                      new Position(pos, annoBeginTime, annoEndTime));
					annotations.add(a);
				}
			}
		}
		
		return annotations;
	}

	@Override
	public boolean remAnnotation(Position pos) {
		if (pos == null) {
			// warn
			if (LOG.isLoggable(Level.WARNING)) {
				LOG.warning("Remove annotation: Position is null");
			}
			return false;
		}
		
		TierImpl destTier = (TierImpl) transcription.getTierWithId(pos.getTierId());
		
		if (destTier != null) {
			long t = (pos.getBeginTime() + pos.getEndTime()) / 2;
			AbstractAnnotation aa = (AbstractAnnotation) destTier.getAnnotationAtTime(t);
			if (aa != null) {
				// use the undoable delete command
				Command command = ELANCommandFactory.createCommand(transcription, 
						ELANCommandFactory.DELETE_ANNOTATION);
				command.execute(destTier, new Object[]{ELANCommandFactory.getViewerManager(transcription), aa});
				// notify the analyzer of the removal? 
				//should there be an annotationDeleted method in the api? 
				//lexanMediator.remAnnotation(pos);
				return true;
			} else {
				// warn
				if (LOG.isLoggable(Level.WARNING)) {
					LOG.warning("Remove annotation: there is no annotation at the specified position");
				}
			}
		}

		return false;
	}

	/**
	 * Add the listener to the list.
	 * 
	 * @param listener the suggestion set listener
	 */
	@Override
	public void addSuggestionSetListener(SuggestionSetListener listener) {
		if (listener != null && !suggestionListeners.contains(listener)) {
			suggestionListeners.add(listener);
		}
	}
	
	/**
	 * Removes the specified listener from the list of listeners.
	 * 
	 * @param listener the listener to remove
	 */
	@Override
	public void removeSuggestionSetListener(SuggestionSetListener listener) {
		suggestionListeners.remove(listener);
	}

	/**
	 * Implementation of SuggestionSelectionListener.
	 * Notification of selection of a suggestion set by the user.
	 * 
	 * @param event the suggestion selection event. The source of an event is a suggestion viewer model.
	 */
	@Override // SuggestionSelectionListener
	public void suggestionSelected(SuggestionSelectionEvent event) {
		if (event != null) {
			IGTSuggestionViewerModel viewModel = (IGTSuggestionViewerModel) event.getSource();
			int row = event.getSelectedRow();
			try {
				IGTSuggestionModel sugModel = viewModel.getRowData(row);
				SuggestionSet sugSet = sugModel.getSuggestionSet();
				int recursionLevel = viewModel.getRecursionLevel();
				
				// create annotations via a AnnotationsFromSuggestionSetCommand
				// Apply the suggestions, and analyze recursively if enabled.
				annotationsFromSuggestionSet(sugSet, recursionLevel);
				
				notifyLexanSuggestionSelected(row, sugSet);
				
			} catch (ArrayIndexOutOfBoundsException abe) {
				// TODO show message
			}
		}

		// If a recursive annotation occurs above in annotationsFromSuggestionSet(),
		// it will eventually call showNextSuggestionSet() from newAnnotations() 
		// to show the first one.
		// In that case, the call here will open the second one before the first one is closed.
		// Therefore, there is some trickery in interLinearizeActionRecursiveStarting()
		// to prevent that (sugWindowOpened is not set to false).
		showNextSuggestionSet();
	}

	/**
	 * Callback called when the user does not want to act on this particular
	 * suggestion, but does want to see more subsequent suggestions.
	 */
	@Override // SuggestionSelectionListener
	public void suggestionIgnored(SuggestionSelectionEvent event) {
		if (event != null) {
			IGTSuggestionViewerModel viewModel = (IGTSuggestionViewerModel) event.getSource();
			IGTSuggestionModel sugModel = viewModel.getRowData(0);
			SuggestionSet sugSet = sugModel.getSuggestionSet();
			
			notifyLexanSuggestionIgnored(sugSet);
		}
		
		showNextSuggestionSet();
	}

	/**
	 * Callback called when the user does not want to be presented with any
	 * more suggestions.
	 */
	@Override // SuggestionSelectionListener
	public void suggestionClosed(SuggestionSelectionEvent event) {
        // Get rid of any remaining suggestions
        deferredNewAnnotations.clear();
		
		if (event != null) {
			IGTSuggestionViewerModel viewModel = (IGTSuggestionViewerModel) event.getSource();
			IGTSuggestionModel sugModel = viewModel.getRowData(0);
			SuggestionSet sugSet = sugModel.getSuggestionSet();
			
			notifyLexanSuggestionClosed(sugSet);
		}
	}

	private void showNextSuggestionSet() {
		if (LOG.isLoggable(Level.FINE)) {
			LOG.fine(String.format("showNextSuggestionSet: recursionLevel=%d stack has %d items\n",
				recursionLevel, stack.size()));
		}
        // If there are more suggestions queued up, handle them now.
		List<SuggestionSet> sugSets = null;
		
		// At this point, switch to the deepest recursion level to show the "most recent" suggs.
    	deferredNewAnnotations = stack.peek();
    	// Maintain the invariant that
    	// deferredNewAnnotations == stack.get(recursionLevel)
    	recursionLevel = stack.size() - 1;
        if (deferredNewAnnotations != null) {
            sugSets = deferredNewAnnotations.poll();
        }
        
        if (sugSets != null) {
            notifySuggestionSetDelivered(sugSets, stack.size() - 1);
        } else {
			notifyLexanSuggestionClosed(null);
        }
	}

	/**
	 * TODO - Unimplemented!
	 */
	@Override // TextAnalyzerContext
	public void subscribe(Position sourceP, TextAnalyzer lexiconAnalyzer) {
		// TODO Auto-generated method stub
	}

	/**
	 * TODO - Unimplemented!
	 */
	@Override // TextAnalyzerContext
	public void unSubscribe(Position source, TextAnalyzer lexiconAnalyzer) {
		// TODO Auto-generated method stub
	}

	/**
	 * TODO - Unimplemented!
	 */
	@Override // TextAnalyzerContext
	public void saveState(TextAnalyzer lexiconAnalyzer, String saveToString) {
		// TODO Auto-generated method stub
	}

	/**
	 * TODO - Unimplemented!
	 */
	@Override // TextAnalyzerContext
	public String loadState(TextAnalyzer lexiconAnalyzer) {
		// TODO Auto-generated method stub
		return null;
	}

	public void setRecursive(boolean selected) {
		recursive = selected;
	}

	public void annotationDeleted(Position pos) {
		// TODO Auto-generated method stub
	}

	public void annotationsAdded(Position pos) {
		// TODO Auto-generated method stub
	}
	
	/**
	 * Let interested TextAnalyzers know about what selections were chosen by the user.
	 */
	private List<LexanSuggestionSelectionListener> suggestionSelectionListeners;
	
	@Override // TextAnalyzerContext
	public void addSuggestionSelectionListener(LexanSuggestionSelectionListener l) {
		if (suggestionSelectionListeners == null) {
			suggestionSelectionListeners = new ArrayList<LexanSuggestionSelectionListener>();
		}
		
		if (l != null && !suggestionSelectionListeners.contains(l)) {
			suggestionSelectionListeners.add(l);
		}
	}

	@Override // TextAnalyzerContext
	public void removeSuggestionSelectionListener(LexanSuggestionSelectionListener l) {
		if (suggestionSelectionListeners != null) {
			suggestionSelectionListeners.remove(l);
		}
	}
	
	// Handle LexanSuggestionSelectionListeners:
	private void notifyLexanSuggestionSelected(int nr, SuggestionSet selection) {
		if (suggestionSelectionListeners != null) {
			for (LexanSuggestionSelectionListener l : suggestionSelectionListeners) {
				l.suggestionSelected(nr, selection);
			}
		}
	}

	private void notifyLexanSuggestionIgnored(SuggestionSet sugSet) {
		if (suggestionSelectionListeners != null) {
			for (LexanSuggestionSelectionListener l : suggestionSelectionListeners) {
				l.suggestionIgnored(sugSet);
			}
		}
	}

	private void notifyLexanSuggestionClosed(SuggestionSet sugSet) {
		if (suggestionSelectionListeners != null) {
			for (LexanSuggestionSelectionListener l : suggestionSelectionListeners) {
				l.suggestionClosed(sugSet);
			}
		}

		if (stack != null && stack.size() > 1) {
			if (LOG.isLoggable(Level.FINE)) {
				LOG.fine(String.format("notifyLexanSuggestionClosed: pop stack (had %d items), recursionLevel = %d, show next suggestion set, sugWindowOpened=%s\n",
					stack.size(), recursionLevel, String.valueOf(sugWindowOpened)));
			}
	    	stack.pop();
	    	deferredNewAnnotations = stack.peek();
	    	// restart the GUI on the popped queue.
			sugWindowOpened = true;
	    	showNextSuggestionSet();
		}
	}
}
